<!DOCTYPE html>
<html>
<head>
    <title>{{project}}</title>
    <meta http-equiv="Content-Type" content="text/html;charset=UTF-8">
    <link rel="shortcut icon" href="{{request.application_url}}/assets/media/pyramid-16x16.png" />
</head>
<body>
    {% block body %}
    {% endblock %}
</body>
</html>